(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
  typeof define === 'function' && define.amd ? define(['exports'], factory) :
  (global = global || self, factory(global.Rax = {}));
}(this, function (exports) { 'use strict';

  /*
   * Stateful things in runtime
   */
  var Host = {
    mountID: 1,
    // Current owner component
    owner: null,
    isUpdating: false,
    // Roots
    rootComponents: {},
    rootInstances: {},
    // Inject
    driver: null
  };

  function Element(type, key, ref, props, owner) {
    return {
      // Built-in properties that belong on the element
      type: type,
      key: key,
      ref: ref,
      props: props,
      // Record the component responsible for creating this element.
      _owner: owner
    };
  }

  function traverseChildren(children, result) {
    if (Array.isArray(children)) {
      for (var i = 0, l = children.length; i < l; i++) {
        traverseChildren(children[i], result);
      }
    } else {
      result.push(children);
    }
  }

  function flattenChildren(children) {
    if (children == null) {
      return children;
    }

    var result = [];
    traverseChildren(children, result);

    if (result.length === 1) {
      result = result[0];
    }

    return result;
  }

  var RESERVED_PROPS = {
    key: true,
    ref: true
  };

  function getRenderErrorInfo() {
    var ownerComponent = Host.owner;

    if (ownerComponent) {
      var name = ownerComponent.getName();

      if (name) {
        return ' Check the render method of `' + name + '`.';
      }
    }

    return '';
  }

  function createElement(type, config, children) {
    if (type == null) {
      throw Error('createElement: type should not be null or undefined.' + getRenderErrorInfo());
    } // Reserved names are extracted


    var props = {};
    var propName;
    var key = null;
    var ref = null;
    var ownerComponent = Host.owner;

    if (config != null) {
      ref = config.ref === undefined ? null : config.ref;
      key = config.key === undefined ? null : String(config.key);

      if (typeof ref === 'string' && !ownerComponent) {
        console.warn('createElement: adding a string ref outside the component’s render() method.');
      } // Remaining properties are added to a new props object


      for (propName in config) {
        if (!RESERVED_PROPS[propName]) {
          props[propName] = config[propName];
        }
      }
    } // Children arguments can be more than one


    var childrenLength = arguments.length - 2;

    if (childrenLength > 0) {
      if (childrenLength === 1 && !Array.isArray(children)) {
        props.children = children;
      } else {
        var childArray = children;

        if (childrenLength > 1) {
          childArray = new Array(childrenLength);

          for (var i = 0; i < childrenLength; i++) {
            childArray[i] = arguments[i + 2];
          }
        }

        props.children = flattenChildren(childArray);
      }
    } // Resolve default props


    if (type && type.defaultProps) {
      var defaultProps = type.defaultProps;

      for (propName in defaultProps) {
        if (props[propName] === undefined) {
          props[propName] = defaultProps[propName];
        }
      }
    }

    return new Element(type, key, ref, props, ownerComponent);
  }

  function _inheritsLoose(subClass, superClass) {
    subClass.prototype = Object.create(superClass.prototype);
    subClass.prototype.constructor = subClass;
    subClass.__proto__ = superClass;
  }

  /**
   * Base component class.
   */
  var Component =
  /*#__PURE__*/
  function () {
    function Component(props, context) {
      this.props = props;
      this.context = context;
      this.refs = {};
    }

    var _proto = Component.prototype;

    _proto.setState = function setState(partialState, callback) {
      // The updater property is injected when composite component mounting
      this.updater.setState(this, partialState, callback);
    };

    _proto.forceUpdate = function forceUpdate(callback) {
      this.updater.forceUpdate(this, callback);
    };

    return Component;
  }();

  var ValueEmitter =
  /*#__PURE__*/
  function () {
    function ValueEmitter(defaultValue) {
      this.handlers = [];
      this.value = defaultValue;
    }

    var _proto = ValueEmitter.prototype;

    _proto.on = function on(handler) {
      this.handlers.push(handler);
    };

    _proto.off = function off(handler) {
      this.handlers = this.handlers.filter(function (h) {
        return h !== handler;
      });
    };

    _proto.emit = function emit() {
      var _this = this;

      this.handlers.forEach(function (handler) {
        return handler(_this.value);
      });
    };

    return ValueEmitter;
  }();

  var uniqueId = 0;
  function createContext(defaultValue) {
    var _Provider$childContex, _Consumer$contextType;

    var contextProp = '__context_' + uniqueId++ + '__';

    var Provider =
    /*#__PURE__*/
    function (_Component) {
      _inheritsLoose(Provider, _Component);

      function Provider() {
        var _this2;

        _this2 = _Component.call(this) || this;
        _this2.emitter = new ValueEmitter(defaultValue);
        return _this2;
      }

      var _proto2 = Provider.prototype;

      _proto2.getChildContext = function getChildContext() {
        var _ref;

        return _ref = {}, _ref[contextProp] = this.emitter, _ref;
      };

      _proto2.componentWillMount = function componentWillMount() {
        if (this.props.value !== undefined) {
          this.emitter.value = this.props.value;
        }
      };

      _proto2.componentWillReceiveProps = function componentWillReceiveProps(nextProps) {
        if (this.props.value !== nextProps.value) {
          this.emitter.value = nextProps.value;
        }
      };

      _proto2.componentDidUpdate = function componentDidUpdate(prevProps) {
        if (this.props.value !== prevProps.value) {
          this.emitter.emit();
        }
      };

      _proto2.render = function render() {
        return this.props.children;
      };

      return Provider;
    }(Component);

    Provider.childContextTypes = (_Provider$childContex = {}, _Provider$childContex[contextProp] = function () {}, _Provider$childContex);
    Provider.contextProp = contextProp;
    Provider.defaultValue = defaultValue;

    var Consumer =
    /*#__PURE__*/
    function (_Component2) {
      _inheritsLoose(Consumer, _Component2);

      function Consumer(props, context) {
        var _this3;

        _this3 = _Component2.call(this, props, context) || this;
        _this3.state = {
          value: _this3.readContext(_this3.context)
        };

        _this3.onUpdate = function (value) {
          return _this3.state.value !== value && _this3.setState({
            value: value
          });
        };

        return _this3;
      }

      var _proto3 = Consumer.prototype;

      _proto3.readContext = function readContext(context) {
        return context[contextProp] ? context[contextProp].value : defaultValue;
      };

      _proto3.componentDidMount = function componentDidMount() {
        if (this.context[contextProp]) {
          this.context[contextProp].on(this.onUpdate);
        }
      };

      _proto3.componentWillReceiveProps = function componentWillReceiveProps(nextProps, nextContext) {
        var newContextValue = this.readContext(nextContext);

        if (this.state.value !== newContextValue) {
          this.setState({
            value: newContextValue
          });
        }
      };

      _proto3.componentWillUnmount = function componentWillUnmount() {
        if (this.context[contextProp]) {
          this.context[contextProp].off(this.onUpdate);
        }
      };

      _proto3.render = function render() {
        var children = this.props.children;
        var consumer = Array.isArray(children) ? children[0] : children;

        if (typeof consumer === 'function') {
          return consumer(this.state.value);
        }
      };

      return Consumer;
    }(Component);

    Consumer.contextTypes = (_Consumer$contextType = {}, _Consumer$contextType[contextProp] = function () {}, _Consumer$contextType);
    return {
      Provider: Provider,
      Consumer: Consumer
    };
  }

  function createRef() {
    return {
      current: null
    };
  }

  var performUpdate = null;

  var setImmediatePolyfill = function setImmediatePolyfill(job) {
    return setTimeout(job, 0);
  }; // 0s


  var scheduleImmediateCallback = typeof setImmediate === 'undefined' ? setImmediatePolyfill : setImmediate;
  var cancelImmediateCallback = typeof clearImmediate === 'undefined' ? clearTimeout : clearImmediate;
  var beforeNextRenderCallbacks = [];
  var beforeNextRenderCallbackId; // Commit before next render

  function commit() {
    var preCallbacks = beforeNextRenderCallbacks;
    beforeNextRenderCallbacks = [];
    preCallbacks.forEach(function (callback) {
      return callback();
    });
    preCallbacks = null; // Exec callback maybe schedule a work

    performUpdate();
  } // Schedule before next render


  var schedule = function schedule(callback) {
    if (beforeNextRenderCallbacks.length === 0) {
      beforeNextRenderCallbackId = scheduleImmediateCallback(commit);
    }

    beforeNextRenderCallbacks.push(callback);
  }; // Flush before next render

  function flush() {
    if (beforeNextRenderCallbacks.length !== 0) {
      cancelImmediateCallback(beforeNextRenderCallbackId);
      commit();
    }
  }
  function setUpdater(handle) {
    performUpdate = handle;
  }

  function getCurrentRenderingInstance() {
    var currentInstance = Host.owner._instance;

    if (currentInstance) {
      return currentInstance;
    } else {
      throw Error('Hooks can only be called inside a component.');
    }
  }

  function areInputsEqual(inputs, prevInputs) {
    for (var i = 0; i < inputs.length; i++) {
      var val1 = inputs[i];
      var val2 = prevInputs[i];

      if (val1 === val2 && (val1 !== 0 || 1 / val1 === 1 / val2) || val1 !== val1 && val2 !== val2 // eslint-disable-line no-self-compare
      ) {
          continue;
        }

      return false;
    }

    return true;
  }

  function useState(initialState) {
    var currentInstance = getCurrentRenderingInstance();
    var hookID = currentInstance.getHookID();
    var hooks = currentInstance.getHooks();

    if (!hooks[hookID]) {
      // state lazy initializer
      if (typeof initialState === 'function') {
        initialState = initialState();
      }

      var setState = function setState(newState) {
        var current = hooks[hookID][0];

        if (typeof newState === 'function') {
          newState = newState(current);
        }

        if (newState !== current) {
          // This is a render phase update.  After this render pass, we'll restart
          if (Host.owner && Host.owner._instance === currentInstance) {
            hooks[hookID][0] = newState;
            currentInstance.isScheduled = true;
          } else {
            !Host.isUpdating && flush();
            hooks[hookID][0] = newState;
            currentInstance.update();
          }
        }
      };

      hooks[hookID] = [initialState, setState];
    }

    return hooks[hookID];
  }
  function useContext(context) {
    var currentInstance = getCurrentRenderingInstance();
    return currentInstance.readContext(context);
  }
  function useEffect(effect, inputs) {
    useEffectImpl(effect, inputs, true);
  }
  function useLayoutEffect(effect, inputs) {
    useEffectImpl(effect, inputs);
  }

  function useEffectImpl(effect, inputs, defered) {
    var currentInstance = getCurrentRenderingInstance();
    var hookID = currentInstance.getHookID();
    var hooks = currentInstance.getHooks();
    inputs = inputs != null ? inputs : [effect];

    if (!hooks[hookID]) {
      var create = function create(immediately) {
        if (!immediately && defered) return schedule(function () {
          return create(true);
        });
        var current = create.current;

        if (current) {
          // Set this to true to prevent re-entrancy
          var previousIsRendering = Host.isUpdating;
          Host.isUpdating = true;
          destory.current = current();
          create.current = null;
          Host.isUpdating = previousIsRendering;
        }
      };

      var destory = function destory(immediately) {
        if (!immediately && defered) return schedule(function () {
          return destory(true);
        });
        var current = destory.current;

        if (current) {
          // Set this to true to prevent re-entrancy
          var previousIsRendering = Host.isUpdating;
          Host.isUpdating = true;
          current();
          destory.current = null;
          Host.isUpdating = previousIsRendering;
        }
      };

      create.current = effect;
      hooks[hookID] = {
        create: create,
        destory: destory,
        prevInputs: inputs,
        inputs: inputs
      };
      currentInstance.didMount.push(create);
      currentInstance.willUnmount.push(destory);
      currentInstance.didUpdate.push(function () {
        var _hooks$hookID = hooks[hookID],
            prevInputs = _hooks$hookID.prevInputs,
            inputs = _hooks$hookID.inputs,
            create = _hooks$hookID.create;

        if (prevInputs == null || !areInputsEqual(inputs, prevInputs)) {
          destory();
          create();
        }
      });
    } else {
      var hook = hooks[hookID];
      var _create = hook.create,
          prevInputs = hook.inputs;
      hook.inputs = inputs;
      hook.prevInputs = prevInputs;
      _create.current = effect;
    }
  }

  function useImperativeHandle(ref, create, inputs) {
    var nextInputs = inputs != null ? inputs.concat([ref]) : [ref, create];
    useLayoutEffect(function () {
      if (typeof ref === 'function') {
        ref(create());
        return function () {
          return ref(null);
        };
      } else if (ref != null) {
        ref.current = create();
        return function () {
          ref.current = null;
        };
      }
    }, nextInputs);
  }
  function useRef(initialValue) {
    var currentInstance = getCurrentRenderingInstance();
    var hookID = currentInstance.getHookID();
    var hooks = currentInstance.getHooks();

    if (!hooks[hookID]) {
      hooks[hookID] = {
        current: initialValue
      };
    }

    return hooks[hookID];
  }
  function useCallback(callback, inputs) {
    return useMemo(function () {
      return callback;
    }, inputs);
  }
  function useMemo(create, inputs) {
    var currentInstance = getCurrentRenderingInstance();
    var hookID = currentInstance.getHookID();
    var hooks = currentInstance.getHooks();

    if (!hooks[hookID]) {
      hooks[hookID] = [create(), inputs];
    } else {
      var hook = hooks[hookID];
      var prevInputs = hook[1];

      if (!areInputsEqual(inputs, prevInputs)) {
        hook[0] = create();
      }
    }

    return hooks[hookID][0];
  }
  function useReducer(reducer, initialState, initialAction) {
    var currentInstance = getCurrentRenderingInstance();
    var hookID = currentInstance.getHookID();
    var hooks = currentInstance.getHooks();

    if (!hooks[hookID]) {
      if (initialAction) {
        initialState = reducer(initialState, initialAction);
      }

      var dispatch = function dispatch(action) {
        var hook = hooks[hookID]; // reducer will get in the next render, before that we add all
        // actions to the queue

        var queue = hook[2]; // This is a render phase update.  After this render pass, we'll restart

        if (Host.owner && Host.owner._instance === currentInstance) {
          queue.push(action);
          currentInstance.isScheduled = true;
        } else {
          !Host.isUpdating && flush();
          queue.push(action);
          currentInstance.update();
        }
      };

      return hooks[hookID] = [initialState, dispatch, []];
    }

    var hook = hooks[hookID];
    var queue = hook[2];
    var next = hook[0];

    for (var i = 0; i < queue.length; i++) {
      next = reducer(next, queue[i]);
    }

    hook[0] = next;
    hook[2] = [];
    return hooks[hookID];
  }

  var hasOwnProperty = Object.prototype.hasOwnProperty;
  /**
   * inlined Object.is polyfill to avoid requiring consumers ship their own
   * https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Object/is
   */

  function is(x, y) {
    // SameValue algorithm
    if (x === y) {
      // Steps 1-5, 7-10
      // Steps 6.b-6.e: +0 != -0
      return x !== 0 || 1 / x === 1 / y;
    } else {
      // Step 6.a: NaN == NaN
      return x !== x && y !== y;
    }
  }
  /**
   * Performs equality by iterating through keys on an object and returning false
   * when any key has values which are not strictly equal between the arguments.
   * Returns true when the values of all keys are strictly equal.
   */


  function shallowEqual(objA, objB) {
    if (is(objA, objB)) {
      return true;
    }

    if (typeof objA !== 'object' || objA === null || typeof objB !== 'object' || objB === null) {
      return false;
    }

    var keysA = Object.keys(objA);
    var keysB = Object.keys(objB);

    if (keysA.length !== keysB.length) {
      return false;
    } // Test for A's keys different from B.


    for (var i = 0; i < keysA.length; i++) {
      if (!hasOwnProperty.call(objB, keysA[i]) || !is(objA[keysA[i]], objB[keysA[i]])) {
        return false;
      }
    }

    return true;
  }

  function memo(type, compare) {
    if (compare === void 0) {
      compare = shallowEqual;
    }

    // Memo could composed
    if (type.compares) {
      type.compares.push(compare);
    } else {
      type.compares = [compare];
    }

    return type;
  }

  function Fragment(props) {
    return props.children;
  }

  /**
   * Base Component
   */

  var BaseComponent =
  /*#__PURE__*/
  function () {
    function BaseComponent(element) {
      this._currentElement = element;
    }

    var _proto = BaseComponent.prototype;

    _proto.initComponent = function initComponent(parent, parentInstance, context) {
      this._parent = parent;
      this._parentInstance = parentInstance;
      this._context = context;
      this._mountID = Host.mountID++;
    };

    _proto.destoryComponent = function destoryComponent() {

      this._currentElement = null;
      this._nativeNode = null;
      this._parent = null;
      this._parentInstance = null;
      this._context = null;

      if (this._instance) {
        this._instance._internal = null;
        this._instance = null;
      }
    };

    _proto.mountComponent = function mountComponent(parent, parentInstance, context, nativeNodeMounter) {
      this.initComponent(parent, parentInstance, context);
      this.mountNativeNode(nativeNodeMounter);

      var instance = {
        _internal: this
      };
      return instance;
    };

    _proto.updateComponent = function updateComponent() {// Noop
    };

    _proto.unmountComponent = function unmountComponent(shouldNotRemoveChild) {
      if (this._nativeNode && !shouldNotRemoveChild) {
        Host.driver.removeChild(this._nativeNode, this._parent);
      }

      this.destoryComponent();
    };

    _proto.getName = function getName() {
      var currentElement = this._currentElement;
      var instance = this._instance;
      var type = currentElement && currentElement.type;
      var constructor = instance && instance.constructor;
      return type && type.displayName || constructor && constructor.displayName || type && type.name || constructor && constructor.name || type || // Native component's name is type
      currentElement;
    };

    _proto.mountNativeNode = function mountNativeNode(nativeNodeMounter) {
      var nativeNode = this.getNativeNode();
      var parent = this._parent;

      if (nativeNodeMounter) {
        nativeNodeMounter(nativeNode, parent);
      } else {
        Host.driver.appendChild(nativeNode, parent);
      }
    };

    _proto.getNativeNode = function getNativeNode() {
      if (this._nativeNode == null) {
        this._nativeNode = this.createNativeNode();
      }

      return this._nativeNode;
    };

    _proto.getPublicInstance = function getPublicInstance() {
      return this.getNativeNode();
    };

    return BaseComponent;
  }();

  /**
   * Empty Component
   */

  var EmptyComponent =
  /*#__PURE__*/
  function (_BaseComponent) {
    _inheritsLoose(EmptyComponent, _BaseComponent);

    function EmptyComponent() {
      return _BaseComponent.apply(this, arguments) || this;
    }

    var _proto = EmptyComponent.prototype;

    _proto.createNativeNode = function createNativeNode() {
      return Host.driver.createEmpty();
    };

    return EmptyComponent;
  }(BaseComponent);

  /*
   * Ref manager
   */
  var Ref = {
    update: function update(prevElement, nextElement, component) {
      var prevRef = prevElement && prevElement.ref || null;
      var nextRef = nextElement && nextElement.ref || null; // Update refs in owner component

      if (prevRef !== nextRef) {
        // Detach prev RenderedElement's ref
        prevRef != null && this.detach(prevElement._owner, prevRef, component); // Attach next RenderedElement's ref

        nextRef != null && this.attach(nextElement._owner, nextRef, component);
      }
    },
    attach: function attach(ownerComponent, ref, component) {
      if (!ownerComponent) {
        throw Error('Multiple version of Rax used in project.');
      }

      var instance = component.getPublicInstance();

      if (typeof ref === 'function') {
        ref(instance);
      } else if (typeof ref === 'object') {
        ref.current = instance;
      } else {
        ownerComponent._instance.refs[ref] = instance;
      }
    },
    detach: function detach(ownerComponent, ref, component) {
      if (typeof ref === 'function') {
        // When the referenced component is unmounted and whenever the ref changes, the old ref will be called with null as an argument.
        ref(null);
      } else {
        // Must match component and ref could detach the ref on owner when A's before ref is B's current ref
        var instance = component.getPublicInstance();

        if (typeof ref === 'object' && ref.current === instance) {
          ref.current = null;
        } else if (ownerComponent._instance.refs[ref] === instance) {
          delete ownerComponent._instance.refs[ref];
        }
      }
    }
  };

  function instantiateComponent(element) {
    var instance;

    if (element === undefined || element === null || element === false || element === true) {
      instance = new Host.Empty();
    } else if (Array.isArray(element)) {
      instance = new Host.Fragment(element);
    } else if (typeof element === 'object' && element.type) {
      // Special case string values
      if (typeof element.type === 'string') {
        instance = new Host.Native(element);
      } else {
        instance = new Host.Composite(element);
      }
    } else if (typeof element === 'string' || typeof element === 'number') {
      instance = new Host.Text(String(element));
    } else {
      throwInvalidComponentError(element);
    }

    instance._mountIndex = 0;
    return instance;
  }

  function throwInvalidComponentError(element) {
    throw Error("Invalid element type: " + element + ". (current: " + (typeof element === 'object' && Object.keys(element) || typeof element) + ")");
  }

  function shouldUpdateComponent(prevElement, nextElement) {
    var prevEmpty = prevElement === null;
    var nextEmpty = nextElement === null;

    if (prevEmpty || nextEmpty) {
      return prevEmpty === nextEmpty;
    }

    var prevType = typeof prevElement;
    var nextType = typeof nextElement;

    if (prevType === 'string' || prevType === 'number') {
      return nextType === 'string' || nextType === 'number';
    } else {
      // prevElement and nextElement could be array, typeof [] is "object"
      return prevType === 'object' && nextType === 'object' && prevElement.type === nextElement.type && prevElement.key === nextElement.key;
    }
  }

  var getElementKeyName = (function (children, element, index) {
    var elementKey = element && element.key;
    var defaultName = '.' + index.toString(36); // Inner child name default format fallback
    // Key should must be string type

    if (typeof elementKey === 'string') {
      var keyName = '$' + elementKey; // Child keys must be unique.

      var keyUnique = children[keyName] === undefined; // Only the first child will be used when encountered two children with the same key

      if (!keyUnique) console.warn("Encountered two children with the same key \"" + elementKey + "\".");
      return keyUnique ? keyName : defaultName;
    } else {
      return defaultName;
    }
  });

  var rootID = 1;

  var Root =
  /*#__PURE__*/
  function (_Component) {
    _inheritsLoose(Root, _Component);

    function Root() {
      var _this;

      _this = _Component.call(this) || this;
      _this.state = {
        // Using fragment instead of null for avoid create a comment node when init mount
        element: []
      };
      _this.rootID = rootID++;
      return _this;
    }

    var _proto = Root.prototype;

    _proto.getPublicInstance = function getPublicInstance() {
      return this.getRenderedComponent().getPublicInstance();
    };

    _proto.getRenderedComponent = function getRenderedComponent() {
      return this._internal._renderedComponent;
    };

    _proto.update = function update(element) {
      this.setState({
        element: element
      });
    };

    _proto.render = function render() {
      return this.state.element;
    };

    return Root;
  }(Component);

  /**
   * Instance manager
   */

  var KEY = '__r';
  var Instance = {
    set: function set(node, instance) {
      if (!node[KEY]) {
        node[KEY] = instance; // Record root instance to roots map

        if (instance.rootID) {
          Host.rootInstances[instance.rootID] = instance;
          Host.rootComponents[instance.rootID] = instance._internal;
        }
      }
    },
    get: function get(node) {
      return node[KEY];
    },
    remove: function remove(node) {
      var instance = this.get(node);

      if (instance) {
        node[KEY] = null;

        if (instance.rootID) {
          delete Host.rootComponents[instance.rootID];
          delete Host.rootInstances[instance.rootID];
        }
      }
    },
    mount: function mount(element, container, parentInstance) {

      var driver = Host.driver; // Before render callback

      driver.beforeRender && driver.beforeRender(); // Real native root node is body

      if (container == null) {
        container = driver.createBody();
      } // Get the context from the conceptual parent component.


      var parentContext;

      if (parentInstance) {
        var parentInternal = parentInstance._internal;
        parentContext = parentInternal._processChildContext(parentInternal._context);
      } // Update root component


      var prevRootInstance = this.get(container);

      if (prevRootInstance && prevRootInstance.rootID) {
        if (parentContext) {
          // Using _penddingContext to pass new context
          prevRootInstance._internal._penddingContext = parentContext;
        }

        prevRootInstance.update(element);
        return prevRootInstance;
      } // Init root component with empty children


      var renderedComponent = instantiateComponent(createElement(Root));
      var defaultContext = parentContext || {};
      var rootInstance = renderedComponent.mountComponent(container, null, defaultContext);
      this.set(container, rootInstance); // Mount new element through update queue avoid when there is in rendering phase

      rootInstance.update(element); // After render callback

      driver.afterRender && driver.afterRender(rootInstance);

      return rootInstance;
    }
  };

  function toArray(obj) {
    if (!Array.isArray(obj)) {
      obj = [obj];
    }

    return obj;
  }

  var STYLE = 'style';
  var CHILDREN = 'children';
  var TREE = 'tree';
  var EVENT_PREFIX_REGEXP = /^on[A-Z]/;
  /**
   * Native Component
   */

  var NativeComponent =
  /*#__PURE__*/
  function (_BaseComponent) {
    _inheritsLoose(NativeComponent, _BaseComponent);

    function NativeComponent() {
      return _BaseComponent.apply(this, arguments) || this;
    }

    var _proto = NativeComponent.prototype;

    _proto.mountComponent = function mountComponent(parent, parentInstance, context, nativeNodeMounter) {
      var _this = this;

      this.initComponent(parent, parentInstance, context);
      var currentElement = this._currentElement;
      var props = currentElement.props;
      var type = currentElement.type;
      var children = props.children;
      var appendType = props.append || TREE; // Default is tree
      // Clone a copy for style diff

      this._prevStyleCopy = Object.assign({}, props.style);
      var instance = {
        _internal: this,
        type: type,
        props: props
      };
      this._instance = instance;

      var mountChildren = function mountChildren() {
        if (children != null) {
          _this.mountChildren(children, context);
        }
      };

      if (appendType === TREE) {
        // Should after process children when mount by tree mode
        mountChildren();
        this.mountNativeNode(nativeNodeMounter);
      } else {
        // Should before process children when mount by node mode
        this.mountNativeNode(nativeNodeMounter);
        mountChildren();
      } // Ref acttach


      if (currentElement && currentElement.ref) {
        Ref.attach(currentElement._owner, currentElement.ref, this);
      }

      return instance;
    };

    _proto.mountChildren = function mountChildren(children, context) {
      children = toArray(children);
      var nativeNode = this.getNativeNode();
      return this._mountChildren(nativeNode, children, context);
    };

    _proto._mountChildren = function _mountChildren(parent, children, context, nativeNodeMounter) {
      var _this2 = this;

      var renderedChildren = this._renderedChildren = {};
      var renderedChildrenImage = children.map(function (element, index) {
        var renderedChild = instantiateComponent(element);
        var name = getElementKeyName(renderedChildren, element, index);
        renderedChildren[name] = renderedChild;
        renderedChild._mountIndex = index; // Mount children

        var mountImage = renderedChild.mountComponent(parent, _this2._instance, context, nativeNodeMounter);
        return mountImage;
      });
      return renderedChildrenImage;
    };

    _proto.unmountChildren = function unmountChildren(shouldNotRemoveChild) {
      var renderedChildren = this._renderedChildren;

      if (renderedChildren) {
        for (var name in renderedChildren) {
          var renderedChild = renderedChildren[name];
          renderedChild.unmountComponent(shouldNotRemoveChild);
        }

        this._renderedChildren = null;
      }
    };

    _proto.unmountComponent = function unmountComponent(shouldNotRemoveChild) {
      if (this._nativeNode) {
        var ref = this._currentElement.ref;

        if (ref) {
          Ref.detach(this._currentElement._owner, ref, this);
        }

        Instance.remove(this._nativeNode);

        if (!shouldNotRemoveChild) {
          Host.driver.removeChild(this._nativeNode, this._parent);
        }
      }

      this.unmountChildren(shouldNotRemoveChild);
      this._prevStyleCopy = null;
      this.destoryComponent();
    };

    _proto.updateComponent = function updateComponent(prevElement, nextElement, prevContext, nextContext) {
      // Replace current element
      this._currentElement = nextElement;
      Ref.update(prevElement, nextElement, this);
      var prevProps = prevElement.props;
      var nextProps = nextElement.props;
      this.updateProperties(prevProps, nextProps);
      this.updateChildren(nextProps.children, nextContext);
    };

    _proto.updateProperties = function updateProperties(prevProps, nextProps) {
      var propKey;
      var styleName;
      var styleUpdates;
      var driver = Host.driver;
      var nativeNode = this.getNativeNode();

      for (propKey in prevProps) {
        // Continue children and null value prop or nextProps has some propKey that do noting
        if (propKey === CHILDREN || prevProps[propKey] == null || // Use hasOwnProperty here for avoid propKey name is some with method name in object proptotype
        nextProps.hasOwnProperty(propKey)) {
          continue;
        }

        if (propKey === STYLE) {
          // Remove all style
          var lastStyle = this._prevStyleCopy;

          for (styleName in lastStyle) {
            styleUpdates = styleUpdates || {};
            styleUpdates[styleName] = '';
          }

          this._prevStyleCopy = null;
        } else if (EVENT_PREFIX_REGEXP.test(propKey)) {
          // Remove event
          var eventListener = prevProps[propKey];

          if (typeof eventListener === 'function') {
            driver.removeEventListener(nativeNode, propKey.slice(2).toLowerCase(), eventListener);
          }
        } else {
          // Remove attribute
          driver.removeAttribute(nativeNode, propKey, prevProps[propKey]);
        }
      }

      for (propKey in nextProps) {
        var nextProp = nextProps[propKey];
        var prevProp = propKey === STYLE ? this._prevStyleCopy : prevProps != null ? prevProps[propKey] : undefined; // Continue children or prevProp equal nextProp

        if (propKey === CHILDREN || prevProp === nextProp || nextProp == null && prevProp == null) {
          continue;
        } // Update style


        if (propKey === STYLE) {
          if (nextProp) {
            // Clone property
            nextProp = this._prevStyleCopy = Object.assign({}, nextProp);
          } else {
            this._prevStyleCopy = null;
          }

          if (prevProp != null) {
            // Unset styles on `prevProp` but not on `nextProp`.
            for (styleName in prevProp) {
              if (!nextProp || !nextProp[styleName]) {
                styleUpdates = styleUpdates || {};
                styleUpdates[styleName] = '';
              }
            } // Update styles that changed since `prevProp`.


            for (styleName in nextProp) {
              if (prevProp[styleName] !== nextProp[styleName]) {
                styleUpdates = styleUpdates || {};
                styleUpdates[styleName] = nextProp[styleName];
              }
            }
          } else {
            // Assign next prop when prev style is null
            styleUpdates = nextProp;
          }
        } else if (EVENT_PREFIX_REGEXP.test(propKey)) {
          // Update event binding
          var eventName = propKey.slice(2).toLowerCase();

          if (typeof prevProp === 'function') {
            driver.removeEventListener(nativeNode, eventName, prevProp, nextProps);
          }

          if (typeof nextProp === 'function') {
            driver.addEventListener(nativeNode, eventName, nextProp, nextProps);
          }
        } else {
          // Update other property
          if (nextProp != null) {
            driver.setAttribute(nativeNode, propKey, nextProp);
          } else {
            driver.removeAttribute(nativeNode, propKey, prevProps[propKey]);
          }
        }
      }

      if (styleUpdates) {

        driver.setStyles(nativeNode, styleUpdates);
      }
    };

    _proto.updateChildren = function updateChildren(nextChildrenElements, context) {
      // prev rendered children
      var prevChildren = this._renderedChildren;
      var driver = Host.driver;

      if (nextChildrenElements == null && prevChildren == null) {
        return;
      }

      var nextChildren = {};

      if (nextChildrenElements != null) {
        nextChildrenElements = toArray(nextChildrenElements); // Update next children elements

        for (var index = 0, length = nextChildrenElements.length; index < length; index++) {
          var nextElement = nextChildrenElements[index];
          var name = getElementKeyName(nextChildren, nextElement, index);
          var prevChild = prevChildren && prevChildren[name];
          var prevElement = prevChild && prevChild._currentElement;
          var prevContext = prevChild && prevChild._context; // Try to update between the two of some name that has some element type,
          // and move child in next children loop if need

          if (prevChild != null && shouldUpdateComponent(prevElement, nextElement)) {
            if (prevElement !== nextElement || prevContext !== context) {
              // Pass the same context when updating chidren
              prevChild.updateComponent(prevElement, nextElement, context, context);
            }

            nextChildren[name] = prevChild;
          } else {
            // Unmount the prevChild when some name with nextChild but different element type,
            // and move child node in next children loop
            if (prevChild) {
              prevChild._unmount = true;
            } // The child must be instantiated before it's mounted.


            nextChildren[name] = instantiateComponent(nextElement);
          }
        }
      }

      var prevFirstChild;
      var prevFirstNativeNode;
      var shouldUnmountPrevFirstChild; // Unmount children that are no longer present.

      if (prevChildren != null) {
        for (var _name in prevChildren) {
          var _prevChild = prevChildren[_name];
          var shouldUnmount = _prevChild._unmount || !nextChildren[_name]; // Store old first child ref for append node ahead and maybe delay remove it

          if (!prevFirstChild) {
            shouldUnmountPrevFirstChild = shouldUnmount;
            prevFirstChild = _prevChild;
            prevFirstNativeNode = prevFirstChild.getNativeNode();

            if (Array.isArray(prevFirstNativeNode)) {
              prevFirstNativeNode = prevFirstNativeNode[0];
            }
          } else if (shouldUnmount) {
            _prevChild.unmountComponent();
          }
        }
      }

      if (nextChildren != null) {
        var insertNodes = function insertNodes(nativeNodes, parent) {
          // The nativeNodes maybe fragment, so convert to array type
          nativeNodes = toArray(nativeNodes);

          if (lastPlacedNode) {
            // Should reverse order when insert new child after lastPlacedNode:
            // [lastPlacedNode, *newChild1, *newChild2]
            for (var i = nativeNodes.length - 1; i >= 0; i--) {
              driver.insertAfter(nativeNodes[i], lastPlacedNode);
            }
          } else if (prevFirstNativeNode) {
            // [*newChild1, *newChild2, prevFirstNativeNode]
            for (var _i = 0; _i < nativeNodes.length; _i++) {
              driver.insertBefore(nativeNodes[_i], prevFirstNativeNode);
            }
          } else if (parent) {
            // [*newChild1, *newChild2]
            for (var _i2 = 0; _i2 < nativeNodes.length; _i2++) {
              driver.appendChild(nativeNodes[_i2], parent);
            }
          }
        };

        // `nextIndex` will increment for each child in `nextChildren`
        var nextIndex = 0;
        var lastPlacedNode = null;
        var nextNativeNode = [];

        for (var _name2 in nextChildren) {
          var nextChild = nextChildren[_name2];

          var _prevChild2 = prevChildren && prevChildren[_name2]; // Try to move the some key prevChild but current not at the some position


          if (_prevChild2 === nextChild) {
            var prevChildNativeNode = _prevChild2.getNativeNode();

            if (_prevChild2._mountIndex !== nextIndex) {
              insertNodes(prevChildNativeNode);
            }
          } else {
            // Mount nextChild that in prevChildren there has no some name
            var parent = this.getNativeNode(); // Fragment extended native component, so if parent is fragment should get this._parent

            if (Array.isArray(parent)) {
              parent = this._parent;
            }

            nextChild.mountComponent(parent, this._instance, context, insertNodes // Insert nodes mounter
            );
          } // Update to the latest mount order


          nextChild._mountIndex = nextIndex++; // Get the last child

          lastPlacedNode = nextChild.getNativeNode(); // Push to nextNativeNode

          nextNativeNode = nextNativeNode.concat(lastPlacedNode);

          if (Array.isArray(lastPlacedNode)) {
            lastPlacedNode = lastPlacedNode[lastPlacedNode.length - 1];
          }
        } // Sync update native refs


        if (Array.isArray(this._nativeNode)) {
          // Clear all and push the new array
          this._nativeNode.splice(0, this._nativeNode.length);

          for (var i = 0; i < nextNativeNode.length; i++) {
            this._nativeNode.push(nextNativeNode[i]);
          }
        }
      }

      if (shouldUnmountPrevFirstChild) {
        prevFirstChild.unmountComponent();
      }

      this._renderedChildren = nextChildren;
    };

    _proto.createNativeNode = function createNativeNode() {
      var nativeNode = Host.driver.createElement(this._instance);
      Instance.set(nativeNode, this._instance);
      return nativeNode;
    };

    return NativeComponent;
  }(BaseComponent);

  /**
   * Text Component
   */

  var TextComponent =
  /*#__PURE__*/
  function (_BaseComponent) {
    _inheritsLoose(TextComponent, _BaseComponent);

    function TextComponent() {
      return _BaseComponent.apply(this, arguments) || this;
    }

    var _proto = TextComponent.prototype;

    _proto.updateComponent = function updateComponent(prevElement, nextElement, context) {
      nextElement = String(nextElement); // If text is some value that do not update even there number 1 and string "1"

      if (prevElement !== nextElement) {
        // Replace current element
        this._currentElement = nextElement;
        Host.driver.updateText(this.getNativeNode(), this._currentElement);
      }
    };

    _proto.createNativeNode = function createNativeNode() {
      return Host.driver.createText(this._currentElement);
    };

    return TextComponent;
  }(BaseComponent);

  var RE_RENDER_LIMIT = 24;
  /**
   * Functional Reactive Component Class Wrapper
   */

  var ReactiveComponent =
  /*#__PURE__*/
  function (_Component) {
    _inheritsLoose(ReactiveComponent, _Component);

    function ReactiveComponent(pureRender, ref) {
      var _this;

      _this = _Component.call(this) || this; // A pure function

      _this._render = pureRender;
      _this._hookID = 0; // Number of rerenders

      _this._reRenders = 0;
      _this._hooks = {}; // Handles store

      _this.didMount = [];
      _this.didUpdate = [];
      _this.willUnmount = []; // Is render scheduled

      _this.isScheduled = false;

      if (pureRender.forwardRef) {
        _this.prevForwardRef = _this.forwardRef = ref;
      }

      var compares = pureRender.compares;

      if (compares) {
        _this.shouldComponentUpdate = function (nextProps) {
          // Process composed compare
          var arePropsEqual = true; // Compare push in and pop out

          for (var i = compares.length - 1; i > -1; i--) {
            if (arePropsEqual = compares[i](_this.props, nextProps)) {
              break;
            }
          }

          return !arePropsEqual || _this.prevForwardRef !== _this.forwardRef;
        };
      }

      return _this;
    }

    var _proto = ReactiveComponent.prototype;

    _proto.getHooks = function getHooks() {
      return this._hooks;
    };

    _proto.getHookID = function getHookID() {
      return ++this._hookID;
    };

    _proto.readContext = function readContext(context) {
      var _this2 = this;

      var Provider = context.Provider;
      var unmaskContext = this._internal._context;
      var contextProp = Provider.contextProp;
      var contextEmitter = unmaskContext[contextProp];

      if (contextEmitter) {
        var mountId = this._internal._mountID;

        if (!contextEmitter[mountId]) {
          // One context one updater bind
          contextEmitter[mountId] = {};

          var contextUpdater = function contextUpdater(newContext) {
            if (newContext !== contextEmitter[mountId].renderedContext) {
              _this2.update();
            }
          };

          contextEmitter.on(contextUpdater);
          this.willUnmount.push(function () {
            delete contextEmitter[mountId];
            contextEmitter.off(contextUpdater);
          });
        }

        return contextEmitter[mountId].renderedContext = contextEmitter.value;
      }

      return Provider.defaultValue;
    };

    _proto.componentDidMount = function componentDidMount() {
      this.didMount.forEach(function (handler) {
        return handler();
      });
    };

    _proto.componentDidUpdate = function componentDidUpdate() {
      this.didUpdate.forEach(function (handler) {
        return handler();
      });
    };

    _proto.componentWillUnmount = function componentWillUnmount() {
      this.willUnmount.forEach(function (handler) {
        return handler();
      });
    };

    _proto.update = function update() {
      this.forceUpdate();
    };

    _proto.render = function render() {

      this._hookID = 0;
      this._reRenders = 0;
      this.isScheduled = false;

      var children = this._render(this.props, this.forwardRef ? this.forwardRef : this.context);

      while (this.isScheduled) {
        this._reRenders++;

        if (this._reRenders > RE_RENDER_LIMIT) {
          throw Error('Too many re-renders, the number of renders is limited to prevent an infinite loop.');
        }

        this._hookID = 0;
        this.isScheduled = false;
        children = this._render(this.props, this.forwardRef ? this.forwardRef : this.context);
      }

      return children;
    };

    return ReactiveComponent;
  }(Component);

  var dirtyComponents = [];

  function getPendingCallbacks(internal) {
    return internal._pendingCallbacks;
  }

  function setPendingCallbacks(internal, callbacks) {
    return internal._pendingCallbacks = callbacks;
  }

  function getPendingStateQueue(internal) {
    return internal._pendingStateQueue;
  }

  function setPendingStateQueue(internal, partialState) {
    return internal._pendingStateQueue = partialState;
  }

  function enqueueCallback(internal, callback) {
    var callbackQueue = getPendingCallbacks(internal) || setPendingCallbacks(internal, []);
    callbackQueue.push(callback);
  }

  function enqueueState(internal, partialState) {
    var stateQueue = getPendingStateQueue(internal) || setPendingStateQueue(internal, []);
    stateQueue.push(partialState);
  }

  function runCallbacks(callbacks, context) {
    if (callbacks) {
      for (var i = 0; i < callbacks.length; i++) {
        callbacks[i].call(context);
      }
    }
  }

  function runUpdate(component) {
    var internal = component._internal;

    if (!internal) {
      return;
    }

    Host.isUpdating = true; // If updateComponent happens to enqueue any new updates, we
    // shouldn't execute the callbacks until the next render happens, so
    // stash the callbacks first

    var callbacks = getPendingCallbacks(internal);
    setPendingCallbacks(internal, null);
    var prevElement = internal._currentElement;
    var prevUnmaskedContext = internal._context;
    var nextUnmaskedContext = internal._penddingContext || prevUnmaskedContext;
    internal._penddingContext = undefined;

    if (getPendingStateQueue(internal) || internal._isPendingForceUpdate) {
      internal.updateComponent(prevElement, prevElement, prevUnmaskedContext, nextUnmaskedContext);
    }

    runCallbacks(callbacks, component);
    Host.isUpdating = false;
  }

  function mountOrderComparator(c1, c2) {
    return c2._internal._mountID - c1._internal._mountID;
  }

  function performUpdate$1() {
    if (Host.isUpdating) {
      return;
    }

    var dirties;
    var component;

    while (dirtyComponents.length > 0) {
      // Before next render, we will flush all the PassiveEffects
      flush(); // Since reconciling a component higher in the owner hierarchy usually (not
      // always -- see shouldComponentUpdate()) will reconcile children, reconcile
      // them before their children by sorting the array.

      dirties = dirtyComponents.sort(mountOrderComparator); // Reset dirtyComponents, using to collect dirtyComponent
      // generated by the next render

      dirtyComponents = [];

      while (component = dirties.pop()) {
        runUpdate(component);
      }
    }
  }

  function scheduleUpdate(component) {
    if (dirtyComponents.indexOf(component) < 0) {
      dirtyComponents.push(component);
    }

    performUpdate$1();
  }

  function requestUpdate(component, partialState, callback) {
    var internal = component._internal;

    if (!internal) {
      return;
    }

    !Host.isUpdating && flush();

    if (callback) {
      enqueueCallback(internal, callback);
    }

    var hasComponentRendered = internal._renderedComponent; // setState

    if (partialState) {
      enqueueState(internal, partialState); // State pending when request update in componentWillMount and componentWillReceiveProps,
      // isPendingState default is false value (false or null) and set to true after componentWillReceiveProps,
      // _renderedComponent is null when componentWillMount exec.

      if (!internal._isPendingState && hasComponentRendered) {
        scheduleUpdate(component);
      }
    } else {
      // forceUpdate
      internal._isPendingForceUpdate = true;

      if (hasComponentRendered) {
        scheduleUpdate(component);
      }
    }
  }

  setUpdater(performUpdate$1);
  var Updater = {
    setState: requestUpdate,
    forceUpdate: function forceUpdate(component, callback) {
      requestUpdate(component, null, callback);
    },
    runCallbacks: runCallbacks
  };

  function performInSandbox(fn, instance, callback) {
    try {
      return fn();
    } catch (e) {
      if (callback) {
        callback(e);
      } else {
        handleError(instance, e);
      }
    }
  }

  function handleError(instance, error) {
    var boundary;

    while (instance) {
      var internal = instance._internal;

      if (typeof instance.componentDidCatch === 'function') {
        boundary = instance;
        break;
      } else if (internal && internal._parentInstance) {
        instance = internal._parentInstance;
      } else {
        break;
      }
    }

    if (boundary) {
      // Should not attempt to recover an unmounting error boundary
      var boundaryInternal = boundary._internal;

      if (boundaryInternal) {
        var callbackQueue = boundaryInternal._pendingCallbacks || (boundaryInternal._pendingCallbacks = []);
        callbackQueue.push(function () {
          return boundary.componentDidCatch(error);
        });
      }
    } else {
      // Do not break when error happens
      setTimeout(function () {
        throw error;
      }, 0);
    }
  }
  /**
   * Composite Component
   */


  var CompositeComponent =
  /*#__PURE__*/
  function (_BaseComponent) {
    _inheritsLoose(CompositeComponent, _BaseComponent);

    function CompositeComponent() {
      return _BaseComponent.apply(this, arguments) || this;
    }

    var _proto = CompositeComponent.prototype;

    _proto.mountComponent = function mountComponent(parent, parentInstance, context, nativeNodeMounter) {

      this.initComponent(parent, parentInstance, context);
      this._updateCount = 0;

      var currentElement = this._currentElement;
      var Component = currentElement.type;
      var ref = currentElement.ref;
      var publicProps = currentElement.props;
      var componentPrototype = Component.prototype; // Context process

      var publicContext = this._processContext(context); // Initialize the public class


      var instance;
      var renderedElement;

      try {
        if (componentPrototype && componentPrototype.render) {
          // Class Component instance
          instance = new Component(publicProps, publicContext);
        } else if (typeof Component === 'function') {
          // Functional reactive component with hooks
          instance = new ReactiveComponent(Component, ref);
        } else {
          throwInvalidComponentError(Component);
        }
      } catch (e) {
        return handleError(parentInstance, e);
      } // These should be set up in the constructor, but as a convenience for
      // simpler class abstractions, we set them up after the fact.


      instance.props = publicProps;
      instance.context = publicContext;
      instance.refs = {}; // Inject the updater into instance

      instance.updater = Updater;
      instance._internal = this;
      this._instance = instance; // Init state, must be set to an object or null

      var initialState = instance.state;

      if (initialState === undefined) {
        // TODO clone the state?
        instance.state = initialState = null;
      }

      var error = null;

      var errorCallback = function errorCallback(e) {
        error = e;
      };

      if (instance.componentWillMount) {
        performInSandbox(function () {
          {
            instance.componentWillMount();
          }
        }, instance, errorCallback);
      }

      if (renderedElement == null) {
        Host.owner = this; // Process pending state when call setState in componentWillMount

        instance.state = this._processPendingState(publicProps, publicContext);
        performInSandbox(function () {
          {
            renderedElement = instance.render();
          }
        }, instance, errorCallback);
        Host.owner = null;
      }

      this._renderedComponent = instantiateComponent(renderedElement);

      this._renderedComponent.mountComponent(this._parent, instance, this._processChildContext(context), nativeNodeMounter);

      if (error) {
        handleError(instance, error);
      }

      if (!currentElement.type.forwardRef && ref) {
        Ref.attach(currentElement._owner, ref, this);
      }

      if (instance.componentDidMount) {
        performInSandbox(function () {
          {
            instance.componentDidMount();
          }
        }, instance);
      } // Trigger setState callback in componentWillMount or boundary callback after rendered


      var callbacks = this._pendingCallbacks;

      if (callbacks) {
        this._pendingCallbacks = null;
        Updater.runCallbacks(callbacks, instance);
      }

      return instance;
    };

    _proto.unmountComponent = function unmountComponent(shouldNotRemoveChild) {
      var instance = this._instance; // Unmounting a composite component maybe not complete mounted
      // when throw error in component constructor stage

      if (instance && instance.componentWillUnmount) {
        performInSandbox(function () {
          instance.componentWillUnmount();
        }, instance);
      }

      if (this._renderedComponent != null) {
        var currentElement = this._currentElement;
        var ref = currentElement.ref;

        if (!currentElement.type.forwardRef && ref) {
          Ref.detach(currentElement._owner, ref, this);
        }

        this._renderedComponent.unmountComponent(shouldNotRemoveChild);

        this._renderedComponent = null;
      } // Reset pending fields
      // Even if this component is scheduled for another async update,
      // it would still be ignored because these fields are reset.


      this._pendingStateQueue = null;
      this._isPendingForceUpdate = false;
      this.destoryComponent();
    }
    /**
     * Filters the context object to only contain keys specified in
     * `contextTypes`
     */
    ;

    _proto._processContext = function _processContext(context) {
      var Component = this._currentElement.type;
      var contextTypes = Component.contextTypes;

      if (!contextTypes) {
        return {};
      }

      var maskedContext = {};

      for (var contextName in contextTypes) {
        maskedContext[contextName] = context[contextName];
      }

      return maskedContext;
    };

    _proto._processChildContext = function _processChildContext(currentContext) {
      var instance = this._instance; // The getChildContext method context should be current instance

      var childContext = instance.getChildContext && instance.getChildContext();

      if (childContext) {
        return Object.assign({}, currentContext, childContext);
      }

      return currentContext;
    };

    _proto._processPendingState = function _processPendingState(props, context) {
      var instance = this._instance;
      var queue = this._pendingStateQueue;

      if (!queue) {
        return instance.state;
      } // Reset pending queue


      this._pendingStateQueue = null;
      var nextState = Object.assign({}, instance.state);

      for (var i = 0; i < queue.length; i++) {
        var partial = queue[i];
        Object.assign(nextState, typeof partial === 'function' ? partial.call(instance, nextState, props, context) : partial);
      }

      return nextState;
    };

    _proto.updateComponent = function updateComponent(prevElement, nextElement, prevUnmaskedContext, nextUnmaskedContext) {
      var instance = this._instance; // Maybe update component that has already been unmounted or failed mount.

      if (!instance) {
        return;
      }

      var willReceive = false;
      var nextContext;
      var nextProps; // Determine if the context has changed or not

      if (this._context === nextUnmaskedContext) {
        nextContext = instance.context;
      } else {
        nextContext = this._processContext(nextUnmaskedContext);
        willReceive = true;
      } // Distinguish between a props update versus a simple state update


      if (prevElement === nextElement) {
        // Skip checking prop types again -- we don't read component.props to avoid
        // warning for DOM component props in this upgrade
        nextProps = nextElement.props;
      } else {
        nextProps = nextElement.props;
        willReceive = true;
      }

      var hasReceived = willReceive && instance.componentWillReceiveProps;

      if (hasReceived) {
        // Calling this.setState() within componentWillReceiveProps will not trigger an additional render.
        this._isPendingState = true;
        performInSandbox(function () {
          instance.componentWillReceiveProps(nextProps, nextContext);
        }, instance);
        this._isPendingState = false;
      } // Update refs


      if (this._currentElement.type.forwardRef) {
        instance.prevForwardRef = prevElement.ref;
        instance.forwardRef = nextElement.ref;
      } else {
        Ref.update(prevElement, nextElement, this);
      } // Shoud update default


      var shouldUpdate = true;
      var prevProps = instance.props;
      var prevState = instance.state; // TODO: could delay execution processPendingState

      var nextState = this._processPendingState(nextProps, nextContext); // ShouldComponentUpdate is not called when forceUpdate is used


      if (!this._isPendingForceUpdate) {
        if (instance.shouldComponentUpdate) {
          shouldUpdate = performInSandbox(function () {
            return instance.shouldComponentUpdate(nextProps, nextState, nextContext);
          }, instance);
        } else if (instance.isPureComponent) {
          // Pure Component
          shouldUpdate = !shallowEqual(prevProps, nextProps) || !shallowEqual(prevState, nextState);
        }
      }

      if (shouldUpdate) {
        this._isPendingForceUpdate = false; // Will set `this.props`, `this.state` and `this.context`.

        var prevContext = instance.context; // Cannot use this.setState() in componentWillUpdate.
        // If need to update state in response to a prop change, use componentWillReceiveProps instead.

        performInSandbox(function () {
          if (instance.componentWillUpdate) {
            instance.componentWillUpdate(nextProps, nextState, nextContext);
          }
        }, instance); // Replace with next

        this._currentElement = nextElement;
        this._context = nextUnmaskedContext;
        instance.props = nextProps;
        instance.state = nextState;
        instance.context = nextContext;

        this._updateRenderedComponent(nextUnmaskedContext);

        performInSandbox(function () {
          if (instance.componentDidUpdate) {
            instance.componentDidUpdate(prevProps, prevState, prevContext);
          }
        }, instance);
        this._updateCount++;
      } else {
        // If it's determined that a component should not update, we still want
        // to set props and state but we shortcut the rest of the update.
        this._currentElement = nextElement;
        this._context = nextUnmaskedContext;
        instance.props = nextProps;
        instance.state = nextState;
        instance.context = nextContext;
      } // Flush setState callbacks set in componentWillReceiveProps or boundary callback


      var callbacks = this._pendingCallbacks;

      if (callbacks) {
        this._pendingCallbacks = null;
        Updater.runCallbacks(callbacks, instance);
      }
    }
    /**
     * Call the component's `render` method and update the DOM accordingly.
     */
    ;

    _proto._updateRenderedComponent = function _updateRenderedComponent(context) {

      var prevRenderedComponent = this._renderedComponent;
      var prevRenderedElement = prevRenderedComponent._currentElement;
      var instance = this._instance;
      var nextRenderedElement;
      Host.owner = this;
      performInSandbox(function () {
        {
          nextRenderedElement = instance.render();
        }
      }, instance);
      Host.owner = null;

      if (shouldUpdateComponent(prevRenderedElement, nextRenderedElement)) {
        var prevRenderedUnmaskedContext = prevRenderedComponent._context;

        var nextRenderedUnmaskedContext = this._processChildContext(context);

        if (prevRenderedElement !== nextRenderedElement || prevRenderedUnmaskedContext !== nextRenderedUnmaskedContext) {
          prevRenderedComponent.updateComponent(prevRenderedElement, nextRenderedElement, prevRenderedUnmaskedContext, nextRenderedUnmaskedContext);
        }
      } else {
        var prevNativeNode = prevRenderedComponent.getNativeNode();
        prevRenderedComponent.unmountComponent(true);
        this._renderedComponent = instantiateComponent(nextRenderedElement);

        this._renderedComponent.mountComponent(this._parent, instance, this._processChildContext(context), function (newNativeNode, parent) {
          prevNativeNode = toArray(prevNativeNode);
          newNativeNode = toArray(newNativeNode);
          var driver = Host.driver; // If the new length large then prev

          var lastNativeNode;

          for (var i = 0; i < newNativeNode.length; i++) {
            var nativeNode = newNativeNode[i];

            if (prevNativeNode[i]) {
              driver.replaceChild(nativeNode, prevNativeNode[i]);
            } else if (lastNativeNode) {
              driver.insertAfter(nativeNode, lastNativeNode);
            } else {
              driver.appendChild(nativeNode, parent);
            }

            lastNativeNode = nativeNode;
          } // If the new length less then prev


          if (newNativeNode.length < prevNativeNode.length) {
            for (var _i = newNativeNode.length; _i < prevNativeNode.length; _i++) {
              driver.removeChild(prevNativeNode[_i]);
            }
          }
        });
      }
    };

    _proto.getNativeNode = function getNativeNode() {
      // 123
      var renderedComponent = this._renderedComponent;

      if (renderedComponent) {
        return renderedComponent.getNativeNode();
      }
    };

    _proto.getPublicInstance = function getPublicInstance() {
      var instance = this._instance; // The functional components cannot be given refs

      if (instance instanceof ReactiveComponent) {
        return null;
      }

      return instance;
    };

    return CompositeComponent;
  }(BaseComponent);

  /**
   * Fragment Component
   */

  var FragmentComponent =
  /*#__PURE__*/
  function (_NativeComponent) {
    _inheritsLoose(FragmentComponent, _NativeComponent);

    function FragmentComponent() {
      return _NativeComponent.apply(this, arguments) || this;
    }

    var _proto = FragmentComponent.prototype;

    _proto.mountComponent = function mountComponent(parent, parentInstance, context, nativeNodeMounter) {
      this.initComponent(parent, parentInstance, context);
      var instance = {
        _internal: this
      };
      this._instance = instance; // Mount children

      this.mountChildren(this._currentElement, context);
      var fragment = this.getNativeNode();

      if (nativeNodeMounter) {
        nativeNodeMounter(fragment, parent);
      } else {
        for (var i = 0; i < fragment.length; i++) {
          Host.driver.appendChild(fragment[i], parent);
        }
      }

      return instance;
    };

    _proto.mountChildren = function mountChildren(children, context) {
      var fragment = this.getNativeNode();
      return this._mountChildren(this._parent, children, context, function (nativeNode) {
        nativeNode = toArray(nativeNode);

        for (var i = 0; i < nativeNode.length; i++) {
          fragment.push(nativeNode[i]);
        }
      });
    };

    _proto.unmountComponent = function unmountComponent(shouldNotRemoveChild) {
      var nativeNode = this._nativeNode;

      if (nativeNode) {
        Instance.remove(nativeNode);

        if (!shouldNotRemoveChild) {
          for (var i = 0; i < nativeNode.length; i++) {
            Host.driver.removeChild(nativeNode[i]);
          }
        }
      } // Do not need remove child when their parent is removed


      this.unmountChildren(true);
      this.destoryComponent();
    };

    _proto.updateComponent = function updateComponent(prevElement, nextElement, prevContext, nextContext) {
      // Replace current element
      this._currentElement = nextElement;
      this.updateChildren(this._currentElement, nextContext);
    };

    _proto.createNativeNode = function createNativeNode() {
      return [];
    };

    _proto.getName = function getName() {
      return 'fragment';
    };

    return FragmentComponent;
  }(NativeComponent);

  function inject(_ref) {
    var driver = _ref.driver,
        measurer = _ref.measurer;
    // Inject component class
    Host.Empty = EmptyComponent;
    Host.Native = NativeComponent;
    Host.Text = TextComponent;
    Host.Fragment = FragmentComponent;
    Host.Composite = CompositeComponent; // Inject render driver

    Host.driver = driver || Host.driver;

    if (!Host.driver) {
      throw Error('Driver not found.');
    }
  }

  function render(element, container, options, callback) {
    // Compatible with `render(element, container, callback)`
    if (typeof options === 'function') {
      callback = options;
      options = null;
    }

    options = options || {}; // Init inject

    inject(options);
    var rootComponent = Instance.mount(element, container, options.parent);
    var componentInstance = rootComponent.getPublicInstance();

    if (callback) {
      callback.call(componentInstance);
    }

    return componentInstance;
  }

  var version = '1.0.0-beta';

  var shared = {
    Host: Host,
    Instance: Instance,
    Element: Element,
    flattenChildren: flattenChildren
  };

  exports.createElement = createElement;
  exports.createRef = createRef;
  exports.createContext = createContext;
  exports.useState = useState;
  exports.useContext = useContext;
  exports.useEffect = useEffect;
  exports.useLayoutEffect = useLayoutEffect;
  exports.useRef = useRef;
  exports.useCallback = useCallback;
  exports.useMemo = useMemo;
  exports.useReducer = useReducer;
  exports.useImperativeHandle = useImperativeHandle;
  exports.Fragment = Fragment;
  exports.memo = memo;
  exports.render = render;
  exports.version = version;
  exports.shared = shared;

  Object.defineProperty(exports, '__esModule', { value: true });

}));
//# sourceMappingURL=rax.js.map
