"use strict";

exports.__esModule = true;
exports.useState = useState;
exports.useContext = useContext;
exports.useEffect = useEffect;
exports.useLayoutEffect = useLayoutEffect;
exports.useImperativeHandle = useImperativeHandle;
exports.useRef = useRef;
exports.useCallback = useCallback;
exports.useMemo = useMemo;
exports.useReducer = useReducer;

var _host = _interopRequireDefault(require("./vdom/host"));

var _scheduler = require("./vdom/scheduler");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function getCurrentRenderingInstance() {
  var currentInstance = _host.default.owner._instance;

  if (currentInstance) {
    return currentInstance;
  } else {
    throw Error('Hooks can only be called inside a component.');
  }
}

function areInputsEqual(inputs, prevInputs) {
  for (var i = 0; i < inputs.length; i++) {
    var val1 = inputs[i];
    var val2 = prevInputs[i];

    if (val1 === val2 && (val1 !== 0 || 1 / val1 === 1 / val2) || val1 !== val1 && val2 !== val2 // eslint-disable-line no-self-compare
    ) {
        continue;
      }

    return false;
  }

  return true;
}

function useState(initialState) {
  var currentInstance = getCurrentRenderingInstance();
  var hookID = currentInstance.getHookID();
  var hooks = currentInstance.getHooks();

  if (!hooks[hookID]) {
    // state lazy initializer
    if (typeof initialState === 'function') {
      initialState = initialState();
    }

    var setState = function setState(newState) {
      var current = hooks[hookID][0];

      if (typeof newState === 'function') {
        newState = newState(current);
      }

      if (newState !== current) {
        // This is a render phase update.  After this render pass, we'll restart
        if (_host.default.owner && _host.default.owner._instance === currentInstance) {
          hooks[hookID][0] = newState;
          currentInstance.isScheduled = true;
        } else {
          !_host.default.isUpdating && (0, _scheduler.flush)();
          hooks[hookID][0] = newState;
          currentInstance.update();
        }
      }
    };

    hooks[hookID] = [initialState, setState];
  }

  return hooks[hookID];
}

function useContext(context) {
  var currentInstance = getCurrentRenderingInstance();
  return currentInstance.readContext(context);
}

function useEffect(effect, inputs) {
  useEffectImpl(effect, inputs, true);
}

function useLayoutEffect(effect, inputs) {
  useEffectImpl(effect, inputs);
}

function useEffectImpl(effect, inputs, defered) {
  var currentInstance = getCurrentRenderingInstance();
  var hookID = currentInstance.getHookID();
  var hooks = currentInstance.getHooks();
  inputs = inputs != null ? inputs : [effect];

  if (!hooks[hookID]) {
    var create = function create(immediately) {
      if (!immediately && defered) return (0, _scheduler.schedule)(function () {
        return create(true);
      });
      var current = create.current;

      if (current) {
        // Set this to true to prevent re-entrancy
        var previousIsRendering = _host.default.isUpdating;
        _host.default.isUpdating = true;
        destory.current = current();
        create.current = null;
        _host.default.isUpdating = previousIsRendering;
      }
    };

    var destory = function destory(immediately) {
      if (!immediately && defered) return (0, _scheduler.schedule)(function () {
        return destory(true);
      });
      var current = destory.current;

      if (current) {
        // Set this to true to prevent re-entrancy
        var previousIsRendering = _host.default.isUpdating;
        _host.default.isUpdating = true;
        current();
        destory.current = null;
        _host.default.isUpdating = previousIsRendering;
      }
    };

    create.current = effect;
    hooks[hookID] = {
      create: create,
      destory: destory,
      prevInputs: inputs,
      inputs: inputs
    };
    currentInstance.didMount.push(create);
    currentInstance.willUnmount.push(destory);
    currentInstance.didUpdate.push(function () {
      var _hooks$hookID = hooks[hookID],
          prevInputs = _hooks$hookID.prevInputs,
          inputs = _hooks$hookID.inputs,
          create = _hooks$hookID.create;

      if (prevInputs == null || !areInputsEqual(inputs, prevInputs)) {
        destory();
        create();
      }
    });
  } else {
    var hook = hooks[hookID];
    var _create = hook.create,
        prevInputs = hook.inputs;
    hook.inputs = inputs;
    hook.prevInputs = prevInputs;
    _create.current = effect;
  }
}

function useImperativeHandle(ref, create, inputs) {
  var nextInputs = inputs != null ? inputs.concat([ref]) : [ref, create];
  useLayoutEffect(function () {
    if (typeof ref === 'function') {
      ref(create());
      return function () {
        return ref(null);
      };
    } else if (ref != null) {
      ref.current = create();
      return function () {
        ref.current = null;
      };
    }
  }, nextInputs);
}

function useRef(initialValue) {
  var currentInstance = getCurrentRenderingInstance();
  var hookID = currentInstance.getHookID();
  var hooks = currentInstance.getHooks();

  if (!hooks[hookID]) {
    hooks[hookID] = {
      current: initialValue
    };
  }

  return hooks[hookID];
}

function useCallback(callback, inputs) {
  return useMemo(function () {
    return callback;
  }, inputs);
}

function useMemo(create, inputs) {
  var currentInstance = getCurrentRenderingInstance();
  var hookID = currentInstance.getHookID();
  var hooks = currentInstance.getHooks();

  if (!hooks[hookID]) {
    hooks[hookID] = [create(), inputs];
  } else {
    var hook = hooks[hookID];
    var prevInputs = hook[1];

    if (!areInputsEqual(inputs, prevInputs)) {
      hook[0] = create();
    }
  }

  return hooks[hookID][0];
}

function useReducer(reducer, initialState, initialAction) {
  var currentInstance = getCurrentRenderingInstance();
  var hookID = currentInstance.getHookID();
  var hooks = currentInstance.getHooks();

  if (!hooks[hookID]) {
    if (initialAction) {
      initialState = reducer(initialState, initialAction);
    }

    var dispatch = function dispatch(action) {
      var hook = hooks[hookID]; // reducer will get in the next render, before that we add all
      // actions to the queue

      var queue = hook[2]; // This is a render phase update.  After this render pass, we'll restart

      if (_host.default.owner && _host.default.owner._instance === currentInstance) {
        queue.push(action);
        currentInstance.isScheduled = true;
      } else {
        !_host.default.isUpdating && (0, _scheduler.flush)();
        queue.push(action);
        currentInstance.update();
      }
    };

    return hooks[hookID] = [initialState, dispatch, []];
  }

  var hook = hooks[hookID];
  var queue = hook[2];
  var next = hook[0];

  for (var i = 0; i < queue.length; i++) {
    next = reducer(next, queue[i]);
  }

  hook[0] = next;
  hook[2] = [];
  return hooks[hookID];
}